/*
 * Copyright (C) 1991 ToM Project
 * All rights reserved.
 *
 * Convert an ordinary file to an object file.
 *
 *	Usage: f2o file_name [object_file_name [symbol_name [size_name]]]
 *
 * Created by Tatsuya Hagino
 */
#include <stdio.h>

#ifndef _WIN32
#include <a.out.h>
#endif

#define ALIGN	4
#ifdef	luna88k
#define DALIGN	16
#endif	luna88k

#ifdef	news5000
#define COFF
#endif	news5000

#define USAGE "Usage: %s file_name [object_file_name [symbol_name [size_name]]]\n"

main(argc, argv)
int argc;
char **argv;
{
  char *program, *filename, *objectname, *symbolname, *sizename;
  FILE *file, *object;
#ifdef	COFF
  struct filehdr filehdr;
  struct aouthdr aouthdr;
  struct scnhdr datahdr;
  struct scnhdr sdatahdr;
#endif	COFF
  struct exec header;
  struct nlist symbol[2];
  char buf[BUFSIZ];
  register int i;
  long size, asize;
  int string_len;

  program = argv[0];
  argc--;
  argv++;
  filename = objectname = symbolname = sizename = NULL;
  if (argc > 0) {
    filename = argv[0];
    argc--;
    argv++;
  }
  else {
    fprintf(stderr, USAGE, program);
    exit(1);
  }
  if (argc > 0) {
    objectname = argv[0];
    argc--;
    argv++;
  }
  else {
    objectname = (char *)malloc(strlen(filename)+3);
    strcpy(objectname, filename);
    strcat(objectname, ".o");
  }
  if (argc > 0) {
    symbolname = argv[0];
    argc--;
    argv++;
  }
  else {
    symbolname = (char *)malloc(strlen(filename)+2);
    strcpy(symbolname, "_");
    strcat(symbolname, filename);
  }

  if (argc > 0) {
    sizename = argv[0];
    argc--;
    argv++;
  }
  else {
    sizename = (char *)malloc(strlen(symbolname)+strlen("_size")+1);
    strcpy(sizename, symbolname);
    strcat(sizename, "_size");
  }

  if (!(file = fopen(filename, "r"))) {
    fprintf(stderr, "%s: cannot read %s\n", program, filename);
    exit(1);
  }
  if (!(object = fopen(objectname, "w"))) {
    fprintf(stderr, "%s: cannot write %s\n", program, objectname);
    exit(1);
  }

  fseek(file, 0L, 2);	/* go to the end of file */
  size = ftell(file);
  asize = (size+(ALIGN-1))& ~(ALIGN-1);
  fseek(file, 0L, 0);	/* rewind */

#ifdef	COFF
  filehdr.f_magic = MIPSBMAGIC;
  filehdr.f_nscns = 2;
  filehdr.f_timdat = time(0);
  filehdr.f_symptr = 0xf4;
  filehdr.f_nsyms = 0x60;
  filehdr.f_opthdr = 0x38;
  filehdr.f_flags = 0;
  aouthdr.magic = OMAGIC;
  aouthdr.vstamp = 0x20B;
  aouthdr.tsize = 0;
  aouthdr.dsize = (asize+sizeof(long)+(ALIGN-1))&~(ALIGN-1);
  aouthdr.bsize = 0;
  aouthdr.entry = 0;
  aouthdr.text_start = 0;
  aouthdr.data_start = 0;
  aouthdr.bss_start = aouthdr.dsize;
  aouthdr.gprmask = 0;
  aouthdr.cprmask[0] = 0;
  aouthdr.cprmask[1] = 0;
  aouthdr.cprmask[2] = 0;
  aouthdr.cprmask[3] = 0;
  aouthdr.gp_value = 0x8010;
  strncpy(datahdr.s_name, ".data", 8);
  datahdr.s_paddr = 0;
  datahdr.s_vaddr = 0;
  datahdr.s_size = (asize+(SCNROUND-1))&~(SCNROUND-1);
  datahdr.s_scnptr = sizeof(filehdr)+sizeof(aouthdr)+sizeof(datahdr)+sizeof(sdatahdr);
  datahdr.s_relptr = 0;
  datahdr.s_lnnptr = 0;
  datahdr.s_nreloc = 0;
  datahdr.s_nlnno = 0;
  datahdr.s_flags = 0x40;
  strncpy(sdatahdr.s_name, ".sdata", 8);
  sdatahdr.s_paddr = datahdr.s_size;
  sdatahdr.s_vaddr = datahdr.s_size;
  sdatahdr.s_size = (sizeof(long)+(SCNROUND-1))&~(SCNROUND-1);
  sdatahdr.s_scnptr = datahdr.s_scnptr+datahdr.s_size;
  sdatahdr.s_relptr = datahdr.s_scnptr+datahdr.s_size+sdatahdr.s_size;
  sdatahdr.s_lnnptr = 0xD4;
  sdatahdr.s_nreloc = 1;
  sdatahdr.s_nlnno = 4;
  sdatahdr.s_flags = 0x200;
  /* GIVE UP! not yet done */
#else  !COFF
  header.a_magic = OMAGIC;
#ifdef	is68k
  header.a_machtype = M_68020;
#endif	is68k
#ifdef	luna88k
  header.a_machtype = M_88000;
#endif	luna88k
#ifdef	sun
  header.a_dynamic = 0;
  header.a_toolversion = TV_SUN4;
  header.a_machtype = M_SPARC;
#endif	sun
  header.a_text = 0;
#ifdef	luna88k
  header.a_data = (asize+sizeof(long)+(DALIGN-1))&~(DALIGN-1);
#else  !luna88k
  header.a_data = asize+sizeof(long);
#endif	luna88k
  header.a_bss = 0;
  header.a_syms = sizeof(struct nlist)*2;
  header.a_entry = 0;
  header.a_trsize = 0;
  header.a_drsize = 0;

  symbol[0].n_un.n_strx = sizeof(long);
  symbol[0].n_type = N_DATA|N_EXT;
  symbol[0].n_other = 0;
  symbol[0].n_desc = 0;
  symbol[0].n_value = 0;
  symbol[1].n_un.n_strx = sizeof(long)+strlen(symbolname)+1;
  symbol[1].n_type = N_DATA|N_EXT;
  symbol[1].n_other = 0;
  symbol[1].n_desc = 0;
  symbol[1].n_value = asize;

  if (fwrite((char *)&header, sizeof(header), 1, object) != 1) {
    fprintf(stderr, "%s: cannot write header\n", program);
    exit(1);
  }
  while ((i = fread(buf, sizeof(char), BUFSIZ, file)) > 0) {
    if (fwrite(buf, sizeof(char), i, object) != i) {
      fprintf(stderr, "%s: cannot write data\n", program);
      exit(1);
    }
  }
  if ((i = asize-size) > 0) {
    while (i-- > 0) fputc(0, object);
  }
  if (fwrite(&size, sizeof(long), 1, object) != 1) {
    fprintf(stderr, "%s: cannot write size data\n", program);
    exit(1);
  }
#ifdef	luna88k
  if (i = (asize+sizeof(long))&(DALIGN-1)) {
	  while (i++ < DALIGN) fputc(0, object);
  }
#endif	luna88k
  if (fwrite((char *)symbol, sizeof(struct nlist), 2, object) != 2) {
    fprintf(stderr, "%s: cannot write symbol\n", program);
    exit(1);
  }
  string_len = sizeof(long)+strlen(symbolname)+1+strlen(sizename)+1;
  if (fwrite((char *)&string_len, sizeof(long), 1, object) != 1) {
    fprintf(stderr, "%s: cannot write string length\n", program);
    exit(1);
  }
  if (fwrite(symbolname, strlen(symbolname)+1, 1, object) != 1) {
    fprintf(stderr, "%s: cannot write string\n", program);
    exit(1);
  }
  if (fwrite(sizename, strlen(sizename)+1, 1, object) != 1) {
    fprintf(stderr, "%s: cannot write string\n", program);
    exit(1);
  }
#endif	COFF

  exit(0);
}
